@@ -1,4 +1,3 @@
-from copy import deepcopy
 from typing import ClassVar
 import datetime
 import logging
@@ -10,7 +9,7 @@ from pydantic._internal._model_construction import ModelMetaclass
 from .utils.supermodel import SuperModel
 from .utils.make_hashable import make_hashable
 from .transaction import transaction
-from .field import Field, JSON
+from .field import Field
 
 
 logger = logging.getLogger("ormantism")
@@ -25,9 +24,12 @@ class _WithSoftDelete(SuperModel):
 class _WithCreatedAtTimestamp(SuperModel):
     created_at: datetime.datetime = None
 
-class _WithTimestamps(_WithCreatedAtTimestamp, _WithSoftDelete):
+class _WithUpdatedAtTimestamp(SuperModel):
     updated_at: datetime.datetime|None = None
 
+class _WithTimestamps(_WithCreatedAtTimestamp, _WithSoftDelete, _WithUpdatedAtTimestamp):
+    pass
+
 class _WithVersion(_WithSoftDelete):
     version: int = 0
 
@@ -37,6 +39,7 @@ class TableMeta(ModelMetaclass):
     def __new__(mcs, name, bases, namespace,
                 with_primary_key: bool=True,
                 with_created_at_timestamp: bool=False,
+                with_updated_at_timestamp: bool=False,
                 with_timestamps: bool=False,
                 versioning_along: tuple[str]=None,
                 connection_name: str=None,
@@ -45,6 +48,8 @@ class TableMeta(ModelMetaclass):
         default_bases: tuple[type[SuperModel]] = tuple()
         if with_primary_key:
             default_bases += (_WithPrimaryKey,)
+        if with_updated_at_timestamp:
+            default_bases += (_WithUpdatedAtTimestamp,)
         if with_created_at_timestamp:
             default_bases += (_WithCreatedAtTimestamp,)
         if with_timestamps:
@@ -97,7 +102,6 @@ class Table(metaclass=TableMeta):
         if self.id is not None and self.id >= 0:
             return
         self.check_read_only(init_data)
-        init_data = self.process_data(init_data)
         # special column for versioning
         if isinstance(self, _WithVersion):
             sql = f"UPDATE {self._get_table_name()} SET deleted_at = CURRENT_TIMESTAMP WHERE deleted_at IS NULL"
@@ -116,17 +120,24 @@ class Table(metaclass=TableMeta):
         # format data
         exclude = set(self.__class__.model_fields)
         include = set()
+        processed_data = self.process_data(init_data)
         formatted_data = {}
-        for name, value in init_data.items():
-            object.__setattr__(self, name, value)
+        # return
+        for name, value in processed_data.items():
+            # object.__setattr__(self, name, value)
             include.add(name)
             if name in exclude:
                 exclude.remove(name)
             else:
                 formatted_data[name] = value
-        formatted_data |= self.model_dump(include=include,
-                                          exclude=exclude,
-                                          mode="json")
+
+        for name, field in self._get_fields().items():
+            if name not in include or name in exclude:
+                continue
+            if field.is_reference:
+                continue
+            formatted_data[name] = field.serialize(getattr(self, name))
+
         # perform insertion
         if formatted_data:
             sql = (f"INSERT INTO {self._get_table_name()} ({", ".join(formatted_data.keys())})\n"
@@ -141,7 +152,40 @@ class Table(metaclass=TableMeta):
         if hasattr(self, "__post_init__"):
             self.__post_init__()
 
-    # INSERT or SELECT
+    # UPDATE
+
+    def on_before_update(self, new_data):
+        """Apply changes to database"""
+        # ensure we're not trying to write read-only fields
+        self.check_read_only(new_data)
+        # fill SET statement
+        set_statement = self.process_data(new_data)
+        if not set_statement:
+            return
+        # fill WHERE statement
+        where_statement = {}
+        if isinstance(self, _WithPrimaryKey):
+            where_statement = {"id": self.id}
+        else:
+            raise NotImplementedError()
+
+        # compute SQL
+        sql = f"UPDATE {self._get_table_name()}\n"
+        sql += f"SET {", ".join(f"{k} = ?" for k in set_statement)}"
+        if isinstance(self, _WithUpdatedAtTimestamp):
+            sql += ", updated_at = CURRENT_TIMESTAMP"
+        sql += f"\nWHERE {"AND ".join(f"{k} = ?" for k in where_statement)}"
+
+        # compute parameters
+        parameters = tuple(value
+                           for statement in (set_statement, where_statement)
+                           for value in statement.values())
+
+        # execute query
+        self._execute_returning(sql=sql,
+                                parameters=parameters)
+
+    # INSERT or SELECT / UPDATE
     @classmethod
     def load_or_create(cls, _search_fields=None, **data):
         # if restriction applies
@@ -152,11 +196,42 @@ class Table(metaclass=TableMeta):
         # return corresponding row if already exists
         loaded = cls.load(**searched_data)
         if loaded:
-            loaded.update(**data)
+            logger.warning(data)
+            changed_data = {key: value
+                            for key, value in data.items()
+                            if getattr(loaded, key) != value}
+            changed_data = {}
+            for name, value in data.items():
+                field = cls._get_field(name)
+                if field.is_reference:
+                    if name not in loaded._lazy_joins:
+                        if value is not None:
+                            changed_data[name] = value.id
+                            raise Exception()
+                    elif value is None:
+                        changed_data[name] = None
+                    else:
+                        foreign_key = loaded._lazy_joins[name]
+                        if isinstance(foreign_key, int):
+                            if foreign_key != value.id:
+                                changed_data[name] = value.id
+                        elif isinstance(foreign_key, tuple) and len(foreign_key) == 2:
+                            if foreign_key[0] != value.__class__ or foreign_key[1] != value.id:
+                                changed_data[name] = value
+                        else:
+                            ValueError("?!")
+
+                elif getattr(loaded, name) != value:
+                    if cls._get_field(name):
+                        changed_data[name] = value
+            logger.warning(changed_data)
+            loaded.update(**changed_data)
             return loaded
         # build new item if not found
         return cls(**data)
 
+    ##
+
     @classmethod
     @cache
     def _has_field(cls, name: str) -> bool:
@@ -227,8 +302,10 @@ class Table(metaclass=TableMeta):
         # create tables for references first
         created.add(cls)
         for field in cls._get_fields().values():
-            if field.is_reference and field.base_type != Table and field.base_type not in created:
-                field.base_type._create_table(created)
+            if field.is_reference:
+                for t in (field.base_type, field.secondary_type):
+                    if issubclass(t, Table) and t != Table and t not in created:
+                        t._create_table(created)
         # initialize statements for table creation
         statements = []
         # id & created_at are special
@@ -238,7 +315,7 @@ class Table(metaclass=TableMeta):
             statements += ["created_at TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP"]
         # other columns are easy
         statements += sum((
-            field.sql_creations
+            list(field.sql_creations)
             for field in cls._get_fields().values()
             if field.name not in ("created_at", "id")
         ), start=[])
@@ -246,7 +323,7 @@ class Table(metaclass=TableMeta):
         statements += [
             f"FOREIGN KEY ({name}_id) REFERENCES {field.base_type._get_table_name()}(id)"
             for name, field in cls._get_fields().items()
-            if field.is_reference
+            if field.is_reference and field.secondary_type is None
             and field.base_type != Table
         ]
         # build & execute SQL
@@ -274,53 +351,38 @@ class Table(metaclass=TableMeta):
             raise AttributeError(f"Cannot set read-only attribute{plural} of {self.__class__.__name__}: {", ".join(read_only_fields)}")
 
     @classmethod
-    def process_data(cls, data: dict) -> dict:
-        data = deepcopy(data)
+    def process_data(cls, data: dict, for_filtering: bool=False) -> dict:
+        data = dict(**data)
         for name in list(data):
+            value = data.pop(name)
             # is there no field for this name?
             try:
                 field = cls._get_field(name)
             except KeyError:
-                continue
+                raise ValueError(f"Invalid key found in data for {cls.__name__}: {name}")
             # so, there is.
             if field.is_reference:
-                # references are special
-                referred: Table = data.pop(name)
-                if field.base_type == Table:
-                    data[f"{field.name}_table"] = referred._get_table_name() if referred else None
-                data[f"{field.name}_id"] = (referred if isinstance(referred, int) else referred.id) if referred else None
+                # scalar reference
+                if field.secondary_type is None:
+                    if field.base_type == Table:
+                        data[f"{field.name}_table"] = value._get_table_name() if value else None
+                    data[f"{field.name}_id"] = (value if isinstance(value, int) else value.id) if value else None
+                # list of references
+                elif isinstance(value, (list, tuple, set)) and issubclass(field.base_type, (list, tuple, set)):
+                    if field.secondary_type == Table:
+                        data[f"{field.name}_tables"] = [referred._get_table_name() for referred in value]
+                    data[f"{field.name}_ids"] = [referred.id for referred in value]
+                # ?
+                else:
+                    raise NotImplementedError(field.name, value, field.base_type, field.secondary_type)
+            # model
+            elif isinstance(value, BaseModel):
+                data[name] = value.model_dump(mode="json")
+            # just some regular stuff
             else:
-                value = data[name]
-                if isinstance(value, BaseModel):
-                    data[name] = value.model_dump(mode="json")
+                data[name] = field.serialize(value, for_filtering=for_filtering)
         return data
 
-    def on_before_update(self, new_data):
-        """Apply changes to database"""
-        # ensure we're not trying to write read-only fields
-        self.check_read_only(new_data)
-        # fill SET statement
-        set_statement = self.process_data(new_data)
-        # fill WHERE statement
-        where_statement = {}
-        if isinstance(self, _WithPrimaryKey):
-            where_statement = {"id": self.id}
-        else:
-            raise NotImplementedError()
-        
-        # compute parameters
-        logger.critical(set_statement)
-        logger.critical(where_statement)
-        parameters = tuple(self._get_field(name).serialize(value) if self._has_field(name) else value
-                           for statement in (set_statement, where_statement)
-                           for name, value in statement.items())
-
-        # execute query
-        self._execute_returning(sql=f"UPDATE {self._get_table_name()}\n"
-                                    f"SET {", ".join(f"{k} = ?" for k in set_statement)}\n"
-                                    f"WHERE {", ".join(f"{k} = ?" for k in where_statement)}",
-                                parameters=parameters)
-
     # DELETE
     def delete(self):
         if isinstance(self, _WithSoftDelete):
@@ -331,9 +393,8 @@ class Table(metaclass=TableMeta):
     # SELECT
     @classmethod
     def load(cls, reversed:bool=True, as_collection:bool=False, with_deleted=False, preload:str|list[str]=None, **criteria) -> "Table":
-        criteria = cls.process_data(criteria)
-        # criteria = {name: cls._get_field(name).serialize(value)
-        #             for name, value in criteria.items()}
+        original_criteria = criteria
+        processed_criteria = cls.process_data(criteria, for_filtering=True)
         if not preload:
             preload = []
         if isinstance(preload, str):
@@ -355,17 +416,24 @@ class Table(metaclass=TableMeta):
         values = []
         sql += "\nWHERE 1 = 1"
         if issubclass(cls, _WithTimestamps) and not with_deleted:
-            criteria = dict(deleted_at=None, **criteria)
-        if criteria:
-            # for name, value in cls.process_data(criteria).items():
-            for name, value in criteria.items():
-                sql += f"\nAND {cls._get_table_name()}.{name}"
+            processed_criteria |= dict(deleted_at=None)
+        if processed_criteria:
+            # for name, value in cls.process_data(processed_criteria).items():
+            for name, value in processed_criteria.items():
+                json_wrap = (cls._has_field(name) and cls._get_field(name).sql_is_json and not isinstance(original_criteria.get(name), str))
+                # column
+                if json_wrap:
+                    sql += f"\nAND JSON({cls._get_table_name()}.{name})"
+                else:
+                    sql += f"\nAND {cls._get_table_name()}.{name}"
+                # comparison
                 if value is None:
                     sql += " IS NULL"
-                elif cls._has_field(name) and cls._get_field(name).column_base_type == JSON:
-                    sql += " = JSON(?)"
                 else:
-                    sql += " = ?"
+                    if json_wrap:
+                        sql += " = JSON(?)"
+                    else:
+                        sql += " = ?"
                     values.append(value)
 
         # ORDER & LIMIT
@@ -432,7 +500,12 @@ class Table(metaclass=TableMeta):
         def lazy_loader(self):
             if not name in self.__dict__:
                 model, identifier = self._lazy_joins[name]
-                value = None if identifier is None else model.load(id=identifier)
+                if issubclass(model, Table):
+                    value = None if identifier is None else model.load(id=identifier)
+                else:
+                    value = [reference_type.load(id=reference_id)
+                             for reference_type, reference_id
+                             in zip(model, identifier)]
                 self.__dict__[name] = value
             return self.__dict__[name]
         setattr(cls, name, property(lazy_loader))
