@@ -1,10 +1,5 @@
 from typing import Iterable
-
-
-def _get_subclasses(base: type) -> Iterable[type]:
-    for subclass in base.__subclasses__()[::-1]:
-        yield from _get_subclasses(subclass)
-        yield subclass
+from .find_subclass import _get_subclasses
 
 
 def get_all_tables() -> Iterable[type["Table"]]:
@@ -15,5 +10,5 @@ def get_all_tables() -> Iterable[type["Table"]]:
 
 def get_table_by_name(name: str) -> type["Table"]:
     for cls in get_all_tables():
-        if cls._get_table_name() == name:
+        if name in (cls._get_table_name(), cls.__name__):
             return cls
